#pragma once
using namespace BWAPI;
using namespace std;

#define ORDER_DELAY 5

#define PYLON_X_COVERAGE 8
#define PYLON_Y_COVERAGE 5

class MacroAIModule : public AIModule {
public:
  virtual void onStart();
  virtual void onEnd(bool isWinner);

  virtual void onSendText(std::string text);
  virtual void onFrame();

  virtual void onUnitCreate(Unit* unit);
  virtual void onUnitDestroy(Unit* unit);
  virtual void onUnitRenegade(Unit* unit);
private:
  static const int gameTimeoutFrame = 25000;
  static const int highestMineralActionPrice = 150;
  static const int maxWaitFrameCount = 300;

  static const int defaultSpeed = 0;
  static const int buildTimeout = 800;

  static const int areaExpandStep = 2;
  static const int restrictedBaseBuildArea = 10;
  static const int restrictedFirstPylonArea = 20;
  
  static const int positionConcealedDistance = 20 * TILE_SIZE;
  static const int strongDefenseDistance = 2;
  static const int weakDefenseDistance = 4;

  Player* player;
  Player* enemyPlayer;

  // units
  list<Unit*> incompleteUnits;

  Unit* baseBuilderProbe;
  Unit* attackBuilderProbe;
  map<int, Unit*> probes;

  Unit* nexus;
  Unit* basePylon;
  Unit* forge;

  map<int, Unit*> attackPylons;
  map<int, Unit*> attackCannons;

  // builder data
  bool baseBuildInProgress;
  set<TilePosition> buildableBaseTiles;
  
  bool attackBuildInProgress;
  set<TilePosition> buildableHostileTiles;

  map<TilePosition, int> stronglyDefendedTiles;
  map<TilePosition, int> weaklyDefendedTiles;

  vector<TilePosition> baseAreaTiles;
  int baseAreaSize;
  bool canBuildBasePylon;
  bool canBuildForge;

  vector<TilePosition> hostileAreaTiles;
  int hostileAreaSize;
  bool canBuildAttackPylon;
  bool canBuildAttackCannon;

  TilePosition startLocationTilePosition;
  Position startLocationPosition;
  Position baseRegionCenter;

  // attack data
  Unit* targetBuilding;
  TilePosition targetTilePosition;
  Position targetRegionCenter;

  bool canSendAttackBuilder;
  bool attackBuilderReady;

  // action data
  int currentFrame;
  int lastActionFrame;
  int lastActionElapsedFrameCount;

  int* lastAvailableActions;
  int lastAvailableActionCount;

  bool isWaiting;
  bool* enabledActions;

  double accumulatedReward;
  RLAgent* agent;
  bool isDraw;

  vector< pair<int, int> > actionHistory;
  vector< vector<double> > stateHistory;

  bool ignore;

  // builder methods
  TilePosition getBasePylonBuildTile();
  TilePosition getForgeBuildTile();
  
  bool isPositionConcealed(Position position);
  bool isPositionDefended(Position position);
  bool isPositionCloseToPylon(Position position);

  int tileDistanceValue(TilePosition tile);

  TilePosition getAttackPylonBuildTile();
  TilePosition getAttackCannonBuildTile();

  // helper methods
  Unit* getNearestMineralToNexus();
  void chooseAttackTarget();

  void manageIncompleteUnits();
  void manageBaseBuilderProbe();
  void manageAttackBuilderProbe();
  void issueOrdersToUnits();

  void fillState();
  
  void manageLostUnit(Unit* unit);
};
