#include <vector>
#include <fstream>
#include <cmath>
#include <cassert>

#include <RL/Matrix.h>

#define ENABLE_ASSERTION 0

Matrix::Matrix(int newRowCount, int newColumnCount)
  : rowCount(newRowCount)
  , columnCount(newColumnCount)
  , size(newRowCount * newColumnCount)
{ 
  this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
  this->values = new double[this->reservedSize];
  this->tempValues = new double[this->reservedSize];
}
	
Matrix::Matrix(int newColumnCount)
  : rowCount(1)
  , columnCount(newColumnCount)
  , size(newColumnCount)
{ 
  this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
  this->values = new double[this->reservedSize];
  this->tempValues = new double[this->reservedSize];
}

Matrix::Matrix()
  : rowCount(1)
  , columnCount(1)
  , size(1)
{ 
  this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
  this->values = new double[this->reservedSize];
  this->tempValues = new double[this->reservedSize];
}

Matrix::Matrix(ifstream* input) {
	*input >> this->rowCount;
	*input >> this->columnCount;

  this->size = this->rowCount * this->columnCount;
  this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);

	this->values = new double[this->reservedSize];
  this->tempValues = new double[this->reservedSize];

	for (int i = 0; i < this->size; i++) {
		
	}

  for (int i = 0; i < this->rowCount; i++) {
		for (int j = 0; j < this->columnCount; j++) {
      *input >> this->values[i + (j * this->rowCount)];
		}
	}
}

double Matrix::get(int row, int column) {
#if ENABLE_ASSERTION
  assert((this->rowCount > row) && (this->columnCount > column));
#endif
	return this->values[row + (column * this->rowCount)];
}

double Matrix::get(int column) {
#if ENABLE_ASSERTION
  assert(this->columnCount > column);
#endif
	return this->values[column];
}

double Matrix::get() {
	return this->values[0];
}

void Matrix::set(int row, int column, double value) {
#if ENABLE_ASSERTION
  assert((this->rowCount > row) && (this->columnCount > column));
#endif
	this->values[row + (column * this->rowCount)] = value;
}

void Matrix::set(int column, double value) {
#if ENABLE_ASSERTION
  assert(this->columnCount > column);
#endif
	this->values[column] = value;
}

void Matrix::set(double value) {
	this->values[0] = value;
}

int Matrix::getRowCount() {
	return this->rowCount;
}

int Matrix::getColumnCount() {
	return this->columnCount;
}

void Matrix::reset() {
  for (int i = 0; i < this->size; i++) {
		this->values[i] = 0;
	}
}

void Matrix::extract(Matrix* matrix) {
  this->rowCount = matrix->rowCount;
  this->columnCount = matrix->columnCount;
  this->size = matrix->size;

  if (this->reservedSize < this->size) {
    this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
    delete[] this->values;
    delete[] this->tempValues;
    this->values = new double[this->reservedSize];
    this->tempValues = new double[this->reservedSize];
  }

  for (int i = 0; i < this->size; i++) {
    this->values[i] = matrix->values[i];
	}
}

void Matrix::transpose() {
  this->columnCount += this->rowCount;
  this->rowCount = this->columnCount - this->rowCount;
  this->columnCount -= this->rowCount;

	for (int i = 0; i < this->rowCount; i++) {
		for (int j = 0; j < this->columnCount; j++) {
			this->tempValues[i + (j * this->rowCount)] = this->values[j + (i * this->columnCount)];
		}
	}
	
  double* swapper = this->values;
	this->values = this->tempValues;
  this->tempValues = swapper;
}

void Matrix::applySigmoid() {
	for (int i = 0; i < this->size; i++) {
		this->values[i] = 1 / (1 + exp(-this->values[i]));
	}
}

void Matrix::applyDiffSigmoid() {
	for (int i = 0; i < this->size; i++) {
		this->values[i] = this->values[i] * (1 - this->values[i]);
	}
}

void Matrix::add(double value) {
	for (int i = 0; i < this->size; i++) {
		this->values[i] += value;
	}
}

void Matrix::add(Matrix* matrix) {
#if ENABLE_ASSERTION
  assert((this->rowCount == matrix->rowCount) && (this->columnCount <= matrix->columnCount));
#endif
	for (int i = 0; i < this->size; i++) {
		this->values[i] += matrix->values[i];
	}
}

void Matrix::subtract(Matrix* matrix) {
#if ENABLE_ASSERTION
  assert((this->rowCount == matrix->rowCount) && (this->columnCount <= matrix->columnCount));
#endif
	for (int i = 0; i < this->size; i++) {
		this->values[i] -= matrix->values[i];
	}
}

void Matrix::multiply(Matrix* matrix) {
#if ENABLE_ASSERTION
  assert(this->columnCount == matrix->rowCount);
#endif
	int newColumnCount = matrix->getColumnCount();
	this->size = this->rowCount * newColumnCount;

  bool expansionRequired = this->reservedSize < this->size;
  if (expansionRequired) {
    this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
    delete[] this->tempValues;
    this->tempValues = new double[this->reservedSize];
  }

	for (int i = 0; i < this->rowCount; i++) {
		for (int j = 0; j < newColumnCount; j++) {
			this->tempValues[i + (j * this->rowCount)] = 0;
			for (int k = 0; k < this->columnCount; k++) {
				this->tempValues[i + (j * this->rowCount)] += this->values[i + (k * this->rowCount)] * matrix->values[k + (j * matrix->rowCount)];
			}
		}
	}

	this->columnCount = newColumnCount;

  if (expansionRequired) {
    delete[] this->values;
    this->values = new double[this->reservedSize];
  }

  double* swapper = this->values;
	this->values = this->tempValues;
  this->tempValues = swapper;
}

void Matrix::multiplyTransposed(Matrix* matrix) {
#if ENABLE_ASSERTION
  assert(this->columnCount == matrix->columnCount);
#endif
	int newColumnCount = matrix->getRowCount();
  this->size = this->rowCount * newColumnCount;

  bool expansionRequired = this->reservedSize < this->size;
  if (expansionRequired) {
    this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);
    delete[] this->tempValues;
    this->tempValues = new double[this->reservedSize];
  }

	for (int i = 0; i < this->rowCount; i++) {
		for (int j = 0; j < newColumnCount; j++) {
			this->tempValues[i + (j * this->rowCount)] = 0;
			for (int k = 0; k < this->columnCount; k++) {
				this->tempValues[i + (j * this->rowCount)] += this->values[i + (k * this->rowCount)] * matrix->values[j + (k * matrix->rowCount)];
			}
		}
	}

	this->columnCount = newColumnCount;

  if (expansionRequired) {
    delete[] this->values;
    this->values = new double[this->reservedSize];
  }

  double* swapper = this->values;
	this->values = this->tempValues;
  this->tempValues = swapper;
}

void Matrix::scalarMultiply(double value) {
	for (int i = 0; i < this->size; i++) {
		this->values[i] *= value;
	}
}

void Matrix::scalarMultiply(Matrix* matrix) {
#if ENABLE_ASSERTION
  assert((this->rowCount == matrix->rowCount) && (this->columnCount <= matrix->columnCount));
#endif
	for (int i = 0; i < this->size; i++) {
		this->values[i] *= matrix->values[i];
	}
}

void Matrix::extend(double value) {
  this->columnCount++;
  int oldSize = this->size;
  this->size = this->rowCount * this->columnCount;

  if (this->reservedSize < this->size) {
    this->reservedSize = MIN_MATRIX_SIZE + (this->size * 2);

    delete[] this->tempValues;
    this->tempValues = new double[this->reservedSize];

    for (int i = 0; i < oldSize; i++) {
      this->tempValues[i] = this->values[i];
    }

    delete[] this->values;    
    this->values = new double[this->reservedSize];
    
    double* swapper = this->values;
	  this->values = this->tempValues;
    this->tempValues = swapper;
  }

  for (int i = 0; i < this->rowCount; i++) {
		this->values[i + ((this->columnCount - 1) * this->rowCount)] = value;
	}
}

void Matrix::save(ofstream* output) {
	*output << this->rowCount << " " << this->columnCount << endl;

	for (int i = 0; i < this->rowCount; i++) {
		for (int j = 0; j < this->columnCount - 1; j++) {
			*output << this->values[i + (j * this->rowCount)] << " ";
		}
		*output << this->values[i + ((this->columnCount - 1) * this->rowCount)] << endl;
	}
}

Matrix::~Matrix() {
	delete[] this->values;
  delete[] this->tempValues;
}
